/**
 * utils
 */

const fs = require('fs');
const pathtool = require('path');
const config = require('../../config');
const chalk = require('chalk');
const _ = require('lodash');
const { DETECT_REGEXP, NO_DEFAULT_REGEXP } = require('./constant');

const printCurrentProject = (appName) => {
  console.log(`当前美杜沙项目： https://mds-portal.alibaba-inc.com/applications/detail?appName=${appName}`);
}

/**
 * 递归删除文件夹下所有文件
 * @param {string} dirPath 绝对路径
 */
function removeDir(dirPath) {
  if (fs.existsSync(dirPath)) {
    const files = fs.readdirSync(dirPath);
    for (let i = 0; i < files.length; i++) {
      const file = files[i];
      const filePath = pathtool.join(dirPath, file);
      if (fs.statSync(filePath).isDirectory()) {
        removeDir(filePath);
      } else {
        fs.unlinkSync(filePath);
      }
    }
    fs.rmdirSync(dirPath);
  }
}

/**
 * 处理命令参数，会跟默认参数进行merge
 * @param {string} command
 * @param {object} params
 */
function processParameters(command, params) {
  // 默认通用配置
  const commonConfig = _.get(config, 'common', {});

  // 每个命令具体的配置
  const commandConfig = _.get(config, command, {});

  // 处理参数值，处理非基本类型的参数值，如数组、对象等
  const processedParams = {};
  Object.keys(params).forEach(key => {
    let value;
    try {
      value = JSON.parse(params[key]);
    } catch (e) {
      value = params[key];
      if (typeof value === 'string' && /^\w+,/.test(value)) {
        value = value.split(',').map(item => item.trim());
      }
    }
    processedParams[key] = value;
  });
  if (processedParams['exportLanguage']) {
    processedParams.supportLocales = processedParams['exportLanguage'];
  }

  return _.assign({}, commonConfig, commandConfig, processedParams);
}

/**
 * 根据指定路径递归扫描文件，提取文案
 *
 * 扫描提取文案应该是一个通用的行为，所以将其放入util内
 * @param {string} dirPath
 * @param {{ bucId, appName, path, extensions }} options
 * @param {any} ig ignore files
 * @returns {{key: string, sourceDefaultMessage: string, targetDefaultMessage: string}[]}
 */
function scanFiles(dirPath, options, ig) {
  const files = fs.readdirSync(dirPath);
  let messages = [];
  for (let i = 0, len = files.length; i < len; i += 1) {
    const file = files[i];
    if (ig && !ig.ignores(file)) {
      const filePath = pathtool.join(dirPath, file);
      const fileExtName = pathtool.extname(filePath);
      const fileStat = fs.lstatSync(filePath);
      if (fileStat.isDirectory()) {
        messages = messages.concat(scanFiles(filePath, options, ig));
      } else if (options.extensions.includes(fileExtName)) {
        messages = messages.concat(processFile(filePath));
      }
    }
  }
  return messages;
}

/**
 * 处理文件，返回文案列表
 * @param {string} path
 */
function processFile(path) {
  // console.log(chalk.cyan('Processing file:', path));
  const content = fs.readFileSync(path, 'utf-8').toString();
  const invalidMessages = getNoDefaultMessages(content, path);
  const messages = extractMessages(content, path).concat(invalidMessages);
  // messages.forEach((message) => console.log(` - key="${message.key}" sourceDefaultMessage="${message.sourceDefaultMessage}" targetDefaultMessage="${message.targetDefaultMessage}"`));
  return messages;
}

/**
 * Given file, extract the i18n message
 * @param {string} content file content
 * @return {array} list of {key, sourceDefaultMessage, targetDefaultMessage}
 */
function extractMessages(content, path) {
  const _messages = [];
  let match;
  while ((match = DETECT_REGEXP.exec(content)) != null) {
    const keyIndex = match[3] ? 3 : 6;
    const messageIndex = match[5] ? 5 : 8;
    let key = match[keyIndex];
    let defaultMessage = match[messageIndex];

    // 仅将`${aaa}`格式的$符号trim掉, '${aaa}'形式不trim
    const shouldTrim = /\.(d|defaultMessage)\([\s\S]*`[\s\S]*\)/.test(match[0]);
    _messages.push({
      key,
      path,
      sourceDefaultMessage: defaultMessage,
      targetDefaultMessage: transformDefaultMessage(defaultMessage, shouldTrim),
    });
  }
  return _messages;
}

/**
 *  校验扫描出Message的合法性
 * @param {{
 *  key: string,
 *  sourceDefaultMessage: string,
 *  targetDefaultMessage: string,
 *  isValid: boolean,
 *  invalidType: string,
 *  path: string,
 * }[]} messages
 */
function verifyMessages(messages) {
  console.log(chalk.magenta('================== 开始校验解析结果 =================='));

  let isOK = true;
  for (let i = 0, len = messages.length; i < len; ++i) {
    let currentMessage = messages[i];
    if (currentMessage.isValid === false && currentMessage.invalidType == 'no_default') {
      isOK = false;
      console.warn(`❌  发现key="${currentMessage.key}"没有默认值 - ${currentMessage.path}`);
      continue;
    }
    const duplicate = _.find(messages, (message) => (currentMessage.key === message.key && currentMessage.sourceDefaultMessage != null && message.sourceDefaultMessage != null && currentMessage.sourceDefaultMessage !== message.sourceDefaultMessage));
    if (duplicate) {
      isOK = false;
      console.warn(`❌  发现key="${currentMessage.key}"默认值不一致："${currentMessage.sourceDefaultMessage}" - ${currentMessage.path}`);
    }
  }
  return isOK;
}

/**
 * 检测代码中使用文案但是没有默认值的case
 * @param {string} content 代码字符串
 * @param {string} filePath 文件路径
 */
function getNoDefaultMessages(content, filePath) {
  const _messages = [];
  let match;
  while ((match = NO_DEFAULT_REGEXP.exec(content)) != null) {
    let key = match[2];
    _messages.push({
      key,
      path: filePath,
      isValid: false,
      invalidType: 'no_default',
    });
  }
  return _messages;
}

/**
 * Transform variables in source defaultMessage (such as ES6 template strings) to target defaultMessage (such as ICU format)
 * Example: 'Hello, ${name}. Welcome to ${where}!' -> 'Hello, {name}. Welcome to {where}!'
 * TODO implement other rules
 * @param {string} message source defaultMessage
 * @return {string} target defaultMessage (for translator)
*/
function transformDefaultMessage(message, shouldTrim) {
  if (!shouldTrim) {
    return message;
  }
  var reg = /\${[a-zA-Z0-9_\s\t]+}/gm;
  var match;
  while ((match = reg.exec(message)) != null) {
    message = message.replace(match[0], match[0].substr(1));
  }
  return message;
}

module.exports = {
  printCurrentProject,
  removeDir,
  processParameters,
  scanFiles,
  verifyMessages,
  getNoDefaultMessages
};
